from __future__ import annotations

from collections import Counter

# io provider
from testscribe.api.io_provider import IOProvider
from testscribe.cli import CLI

g_io: IOProvider = CLI()

# No type hints to avoid circular dependency

# target string as key, Patcher object as value
g_patchers = {}

# set to True during the test generation process
g_test_generating_mode = False

g_test_to_infer_default_inputs = None
g_index_of_test_to_update = -1


def get_initial_mock_name_counter():
    # Make sure there is no mock object named "m" which conflicts with
    # the function m.
    return Counter("m")


# How many times a mock name is shared among mocks
g_mock_name_counter = get_initial_mock_name_counter()
# Mock name to MockProxy instance mapping
g_name_mock_dict = {}



from dataclasses import dataclass
from string import Template
from typing import Any, List, Tuple, Union

from testscribe.code_gen_util import (
    add_indentation,
    spec_contain_param_name_info,
    join_lines,
)
from testscribe.complex_value_util import contain_complex_value
from testscribe.config import g_indent
from testscribe.model_type import MockModel, MockCallModel
from testscribe.result_assertion import generate_assertion
from testscribe.value_util import get_value_repr


@dataclass
class MockCallParams:
    # used for the main assertion. The complex values will be
    # represented by ANY. Additional assertions will follow.
    param_str: str
    # key: parameter name of a complex parameter,  value: the value of that parameter
    name_to_value: dict
    # key: position(index) of a complex parameter in the parameter list
    # value: the value of that parameter.
    # used only when the parameter name is not available
    index_to_value: dict


def get_one_mock_call_arg_str(value_str: str, param_name: str) -> str:
    return f"{param_name}={value_str}" if param_name else value_str


def update_complex_mock_call_arg_map(
    name_to_value: dict,
    index_to_value: dict,
    index: int,
    param_name: str,
    v: Any,
):
    if param_name:
        name_to_value[param_name] = v
    else:
        index_to_value[index] = v
    # return these to make testing easier
    return name_to_value, index_to_value


def get_one_mock_call_param(
    arg_list: List[str],
    name_to_value: dict,
    index_to_value: dict,
    index: int,
    param_name: str,
    v: Any,
):
    if contain_complex_value(v):
        # If the value is too complex such as an object, use ANY here
        # and supplement with additional assertions.
        # todo: treat multi line string value as complex values too?
        value_str = "ANY"
        update_complex_mock_call_arg_map(
            name_to_value=name_to_value,
            index_to_value=index_to_value,
            index=index,
            param_name=param_name,
            v=v,
        )
    else:
        value_str = get_value_repr(v)
    arg_list.append(
        get_one_mock_call_arg_str(value_str=value_str, param_name=param_name)
    )
    # return these to make testing easier
    return arg_list, name_to_value, index_to_value


def get_mock_call_params(param_list: List[Tuple[str, Any]]) -> MockCallParams:
    """
    Generate mock call parameter string and
    dictionaries for complex values to support complex assertions.

    :param param_list: arguments list, the name is the argument name if available
        "" otherwise
    :return:
    """
    # Some functions e.g. some builtin functions may not have the parameter
    # names available.
    arg_list = []
    name_to_value = {}
    index_to_value = {}
    for index, (param_name, v) in enumerate(param_list):
        get_one_mock_call_param(
            arg_list=arg_list,
            name_to_value=name_to_value,
            index_to_value=index_to_value,
            index=index,
            param_name=param_name,
            v=v,
        )
    param_str = ", ".join(arg_list)
    return MockCallParams(
        param_str=param_str, name_to_value=name_to_value, index_to_value=index_to_value
    )


@dataclass
class ComplexMockCallParam:
    # The position of the mock call in the call list
    index: int
    # key: parameter name of a complex parameter,  value: the value of that parameter
    name_to_value: dict
    # key: position(index) of a complex parameter in the parameter list
    # value: the value of that parameter.
    # used only when the parameter name is not available
    index_to_value: dict


@dataclass
class MockCallList:
    # The string to be used in the main statement verifying mock calls
    call_list_str: str
    # Information needed to construct additional assertions for complex
    # parameter values
    complex_params: List[ComplexMockCallParam]


def generate_mock_call_list(m: MockModel) -> MockCallList:
    call_list = []
    complex_params: List[ComplexMockCallParam] = []
    for index, call in enumerate(m.calls):
        call_str, complex_call_param = generate_one_mock_call(index=index, call=call)
        call_list.append(call_str)
        if complex_call_param:
            complex_params.append(complex_call_param)
    return MockCallList(
        call_list_str="\n".join(call_list), complex_params=complex_params
    )


def generate_one_call_str(method_name: str, param_str: str):
    suffix = f".{method_name}" if method_name else ""
    return f"call{suffix}({param_str}),"


def get_one_complex_param(call_params: MockCallParams, index: int):
    if call_params.name_to_value or call_params.index_to_value:
        return ComplexMockCallParam(
            index=index,
            name_to_value=call_params.name_to_value,
            index_to_value=call_params.index_to_value,
        )
    else:
        return None


def generate_one_mock_call(
    index: int, call: MockCallModel
) -> Tuple[str, Union[None, ComplexMockCallParam]]:
    call_params = get_mock_call_params(call.parameters.as_list())
    call_str = generate_one_call_str(
        method_name=call.name, param_str=call_params.param_str
    )
    complex_call_param = get_one_complex_param(call_params=call_params, index=index)
    return call_str, complex_call_param


def get_mock_calls_variable_value(mock_name: str, spec_str: str) -> str:
    return (
        f"get_normalized_mock_calls({mock_name}, {spec_str})"
        if spec_contain_param_name_info(spec_str)
        else f"{mock_name}.mock_calls"
    )


def generate_mock_call_list_verfication(
    mock_name: str, spec_str: str, mock_calls_name: str, call_list_str: str
) -> str:
    """
    Generate the main statement for verifying mock calls for one mock object.

    :param mock_name:
    :param spec_str:
    :param mock_calls_name: the variable name for the actual mock calls
    :param call_list_str: the string representing the call objects
    :return:
    """
    mock_calls_variable_value = get_mock_calls_variable_value(mock_name, spec_str)
    template_str = """\
${mock_calls_name} = ${mock_calls_variable_value}
assert ${mock_calls_name} == [
$call_list_str
]"""
    template = Template(template_str)
    call_list_assertion_str = template.substitute(
        g_indent=g_indent,
        mock_name=mock_name,
        mock_calls_name=mock_calls_name,
        mock_calls_variable_value=mock_calls_variable_value,
        call_list_str=add_indentation(call_list_str, 1),
    )
    return add_indentation(call_list_assertion_str, 1)


def generate_complex_param_verification(
    params: ComplexMockCallParam, mock_calls_name: str
) -> List[str]:
    prefix = f"{mock_calls_name}[{params.index}]"
    keyword_args_statement = [
        generate_assertion(f"{prefix}.kwargs[{repr(k)}]", v)
        for k, v in params.name_to_value.items()
    ]
    position_args_statement = [
        generate_assertion(f"{prefix}.args[{i}]", v)
        for i, v in params.index_to_value.items()
    ]
    return keyword_args_statement + position_args_statement


def generate_mock_call_verification(m: MockModel) -> str:
    call_list = generate_mock_call_list(m)
    if not call_list.call_list_str:
        # assert_not_called doesn't track the child mocks.
        # The mock_calls property does.
        return f"{g_indent}{m.name}.assert_not_called()"
    mock_name = m.name
    mock_calls_name = f"{mock_name}_mock_calls"
    call_list_verification_statement = generate_mock_call_list_verfication(
        mock_name=mock_name,
        spec_str=m.spec_str,
        mock_calls_name=mock_calls_name,
        call_list_str=call_list.call_list_str,
    )
    statements = [call_list_verification_statement]
    for params in call_list.complex_params:
        statements.extend(
            generate_complex_param_verification(
                params=params, mock_calls_name=mock_calls_name
            )
        )
    return join_lines(lines=statements, prepend_new_line=False)


def generate_mock_call_verification_str(mocks: List[MockModel]) -> str:
    statements = [generate_mock_call_verification(m) for m in mocks]
    return join_lines(statements, prepend_new_line=True)



from logging import getLogger, basicConfig
from logging.config import fileConfig
from pathlib import Path
from traceback import extract_stack, format_list
from typing import List, Optional

from testscribe.config import init_config, CONFIG_FILE_NAME, Config
from testscribe.log import log
from testscribe.model_type import TestModel

logger = getLogger(__name__)

LOGGING_CONFIG_FILE_NAME = "test_scribe_logging.conf"
ERROR_RETURN_CODE = 1


def remove_file_if_no_test(file_path: Path, tests: List[TestModel]) -> bool:
    """
    Make sure the target file is deleted if there is no test.

    :param file_path:
    :param tests:
    :return: True if there is no test.
    """
    if tests:
        return False
    if file_path.exists():
        # The missing_ok parameter is added in python 3.8
        file_path.unlink()
        log(f"There is no more tests in the file {file_path}. Delete.")
    return True


def init(config_file_path: Optional[Path] = None) -> Config:
    config_logging()
    if config_file_path is None:
        config_file_path = Path() / CONFIG_FILE_NAME
    config = init_config(config_file_path)
    return config


def does_logging_config_file_exist() -> bool:
    return Path(LOGGING_CONFIG_FILE_NAME).exists()


def config_logging():
    if does_logging_config_file_exist():
        fileConfig(LOGGING_CONFIG_FILE_NAME, disable_existing_loggers=False)
    else:
        basicConfig()


def create_unit_test_file_name(base_name: str):
    return f"test_{base_name}_g.py"


def infer_unit_test_file_path_from_scribe_file(scribe_file_path: Path) -> Path:
    parent = scribe_file_path.parent
    module_name = scribe_file_path.stem
    file_name = create_unit_test_file_name(module_name)
    return parent.joinpath(file_name)


def call_target_function(func, binded_args):
    """
    The function name is used later to find this stack frame.

    :param func:
    :param binded_args:
    :return:
    """
    return func(*binded_args.args, **binded_args.kwargs)


def get_user_code_frames(additional_num_system_frame: int) -> list:
    """
    Return the call stack that belong to the user code being tested.

    Algorithm:
    Search from the bottom of the call stack less the known system frames on top
    looking for the function call_target_function.
    The user call frames start from the call after this frame.

    The number of known system frames on top depend on the caller of this
    function.
    The top most two frames are always the same:
    show_user_call_stack
    get_user_code_frames
    additional_num_system_frame parameter defines the number of additional system
    frames after these two frames.
    """
    frames = extract_stack()
    num_of_system_frames_on_top = 2 + additional_num_system_frame
    i = len(frames) - num_of_system_frames_on_top
    last_system_frame_name = call_target_function.__name__
    while True:
        if frames[i].name == last_system_frame_name:
            last_system_frame_index = i
            break
        else:
            i -= 1
    return frames[last_system_frame_index + 1: -num_of_system_frames_on_top]


def show_user_call_stack(additional_num_system_frame: int) -> None:
    """

    :param additional_num_system_frame: The number of additional system
    ( not user code frames on the top of the stack above the
    get_user_code_frames, show_user_call_stack functions)
    :return:
    """
    user_code_frames = get_user_code_frames(additional_num_system_frame)
    log("Call stack:")
    for frame_str in format_list(user_code_frames):
        log(frame_str)


def get_all_scribe_files(root_path: Path):
    return sorted(root_path.glob("**/*.tscribe"))


def compute_output_root_path(config: Config, output_root_path: Optional[Path]) -> Path:
    if output_root_path:
        return output_root_path
    else:
        return config.output_root_path



import logging
from pathlib import Path
from typing import Optional

from testscribe.execution import save_file
from testscribe.execution_util import (
    init,
    get_all_scribe_files,
    compute_output_root_path,
)
from testscribe.load_scribe_file import load_scribe_file
from testscribe.log import log
from testscribe.model_type import AllTests
from testscribe.test_name import create_test_name

logger = logging.getLogger(__name__)


def regenerate_all_tests(output_root_path: Optional[Path]):
    config = init()
    output_root_path = compute_output_root_path(
        config=config, output_root_path=output_root_path
    )
    regenerate_all_tests_internal(output_root_path)


def regenerate_all_tests_internal(output_root_path: Path):
    log(f"Regenerate all unit tests under {output_root_path}")
    total = 0
    for p in get_all_scribe_files(output_root_path):
        regenerate_tests(p)
        total += 1
    log(f"Regenerated {total} tests.")
    return total


def regenerate_tests(scribe_file_path: Path) -> None:
    init()
    log(f"Regenerate unit tests from {scribe_file_path}")
    all_tests = load_scribe_file(scribe_file_path)
    regenerate_test_names(all_tests)
    save_file(scribe_file_path=scribe_file_path, all_tests=all_tests)


def regenerate_test_names(all_tests: AllTests):
    """
    Regenerate test names from the short names.
    Useful when the target function name changes, and it is used as a test prefix.
    :param all_tests:
    :return:
    """
    tests = all_tests.tests
    for i in range(len(tests) - 1, -1, -1):
        # Start from the last test to keep the test name stable
        # since older tests for the same target appear first.
        t = tests[i]
        t.name = create_test_name(
            all_tests=all_tests,
            index_of_test_to_update=i,
            short_name=t.short_name,
            target_func_name=t.target_func_name,
        )



from typing import Callable, Union, Optional
from unittest.mock import Mock

from testscribe import mocking_support
from testscribe.custom_type import Spec
from testscribe.mock_proxy import MockProxy
from testscribe.user_triggered_exception import UserTriggeredException


def throw(e: BaseException = Exception()):
    """
    Use this function in an input expression to cause a mock call to throw the
    given exception.
    :param e:
    :return:
    """
    return UserTriggeredException(exception=e)


def patch_with_mock(
    target: Union[str, type, Callable], mock_name: str = "", spec: Optional[Spec] = None
) -> None:
    mocking_support.patch_with_mock_internal(
        target=target, mock_name=mock_name, spec=spec
    )


def patch_with_expression(target_str: str, expression: str) -> None:
    mocking_support.patch_with_expression_internal(
        target_str=target_str, expression=expression
    )


def get_normalized_mock_calls(mock: Mock, spec: Spec) -> list:
    """
    Used in the generated unit test code to allow verifying mock calls consistently
    with keyword arguments when parameter names are available.

    :param mock:
    :param spec:
    :return:
    """
    return mocking_support.get_normalized_mock_calls_internal(mock=mock, spec=spec)


def m(spec: Spec, name: str = ""):
    return MockProxy(spec=spec, name=name)



from __future__ import annotations

import logging
from pathlib import Path

from yaml import dump, add_representer

from testscribe.execution_util import remove_file_if_no_test
from testscribe.log import log
from testscribe.model_encoder import encode_model
from testscribe.model_type import AllTests
from testscribe.yaml_tag import add_representer_for_custom_tags

logger = logging.getLogger(__name__)


def generate_scribe_file(scribe_file_path: Path, all_tests: AllTests) -> None:
    if remove_file_if_no_test(file_path=scribe_file_path, tests=all_tests.tests):
        return
    content = encode_model(all_tests)
    # todo: validate the content first before writing to the file
    with scribe_file_path.open(mode="w") as f:
        save_to_yaml(data=content, stream=f)
    log(f"Wrote the generated scribe file to: {scribe_file_path}")


def save_to_yaml(data, stream):
    configure_yaml()
    # Before pyyaml 5.1 the directory keys are sorted. Since 5.1 it is still the default,
    # but the sort_keys parameter was added.
    # It is desirable to control the order so that it is easier to read.
    dump(data=data, stream=stream, sort_keys=False)


def configure_yaml():
    add_representer(str, string_presenter)
    add_representer_for_custom_tags()


def string_presenter(dumper, data):
    """
    Control how the strings are formatted.
    see
    https://yaml-multiline.info/
    https://www.generacodice.com/en/articolo/3292961/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data

    :param dumper:
    :param data:
    :return:
    """
    length = len(data.splitlines())
    if length > 1:
        # only use the block style for multi line strings
        # The flow style is simpler and easier to read for a single line string.
        style = "|"
    else:
        style = None
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style)



from __future__ import annotations

import logging
from typing import Any

from testscribe.code_gen_util import pretty_format_repr_string, add_indentation
from testscribe.complex_value_util import contain_complex_value
from testscribe.config import g_indent
from testscribe.model_type import MockNameModel, ObjectModel, ModuleModel, SetModel

logger = logging.getLogger(__name__)


def generate_class_assertion(target, value: ObjectModel) -> str:
    """
    indentation becomes more complex due to the use of for multiline strings.
    They can't be indented the same way.
    Thus, indentation has to be handled in each special assertion functions.

    :param target:
    :param value:
    :return:
    """
    assert isinstance(value, ObjectModel)
    class_type_assertion = f"{g_indent}assert isinstance({target}, {value.type})"
    representation = value.repr
    if representation:
        member_assertions = [
            generate_assertion(target=f"repr({target})", value=representation)
        ]
    else:
        member_assertions = [
            generate_assertion(f"{target}.{name}", value)
            for name, value in value.members.items()
        ]
    return "\n".join([class_type_assertion] + member_assertions)


def generate_special_collection_assertion(target: str, value) -> str:
    type_assertion = f"{g_indent}assert isinstance({target}, {type(value).__name__})"
    length_assertion = f"{g_indent}assert len({target}) == {len(value)}"
    if isinstance(value, list) or isinstance(value, tuple):
        member_assertions = [
            generate_assertion(f"{target}[{index}]", item)
            for index, item in enumerate(value)
        ]
    else:
        # set that contains complex values should have generated an error earlier.
        # There isn't a reliable way to identify an element in a set
        # to assert separately.
        assert isinstance(value, dict)
        member_assertions = [
            generate_assertion(f"{target}[{repr(key)}]", item)
            for key, item in value.items()
        ]
    return "\n".join([type_assertion, length_assertion] + member_assertions)


def generate_module_assertion(target: str, value: ModuleModel) -> str:
    assert isinstance(value, ModuleModel)
    # Use local import since this is rare
    import_module_type = "from types import ModuleType"
    type_assertion = f"assert type({target}) == ModuleType"
    member_assertion = f"{target}.__name__ == {repr(value.name)}"
    statement = "\n".join([import_module_type, type_assertion, member_assertion])
    return add_indentation(s=statement, level=1)


def generate_set_assertion(target: str, value: SetModel) -> str:
    # Set members are not sorted, use the sorted member list for the generated
    # assertions to keep the result stable.

    # python3 doesn't allow sorting set with mixed types e.g.
    #     s = set([1, 2, 'string'])
    #     print(sorted(list(s)))
    # will result in an error
    #  TypeError: '<' not supported between instances of 'int' and 'str'
    # todo: find a solution. Maybe a switch to disable sorting or catch the exception
    #  and generate a non sorted version?
    assert isinstance(value, SetModel)
    type_assertion = f"{g_indent}assert isinstance({target}, set)"
    member_assertions = generate_assertion(f"sorted(list({target}))", value.elements)
    return "\n".join([type_assertion, member_assertions])


g_complex_type_assertion_function_look_up = {
    ObjectModel: generate_class_assertion,
    ModuleModel: generate_module_assertion,
    SetModel: generate_set_assertion,
}


def generate_complex_type_assertion(target: str, value) -> str:
    func = g_complex_type_assertion_function_look_up.get(
        type(value), generate_special_collection_assertion
    )
    return func(target=target, value=value)


def generate_assertion(target: str, value: Any) -> str:
    if contain_complex_value(value):
        return generate_complex_type_assertion(target, value)
    str_repr = (
        pretty_format_repr_string(value) if isinstance(value, str) else repr(value)
    )
    if (
        value is None
        or isinstance(value, bool)
        # using == to compare a mock object will result in an unwanted __eq__ call.
        # This may mess up the assertion of the mock object's behavior.
        or isinstance(value, MockNameModel)
    ):
        operator = "is"
    else:
        operator = "=="
    return f"{g_indent}assert {target} {operator} {str_repr}"


def generate_result_assertion(result: Any) -> str:
    assertions = generate_assertion("result", result)
    return "\n" + assertions



from __future__ import annotations

import logging.config
from pathlib import Path
from typing import Optional

import typer
from typer import Argument, Option

from testscribe import global_var
from testscribe.create_test_cmd import create_test
from testscribe.delete_test_cmd import delete_test
from testscribe.exception_decorator import exception_handler
from testscribe.move_cmd import move_tests
from testscribe.sync_cmd import regenerate_tests, regenerate_all_tests
from testscribe.update_test_cmd import update_test_cmd

app = typer.Typer(name="testscribe")
logger = logging.getLogger(__name__)


@app.command()
@exception_handler
def create(
    source_file: Path = Argument(
        ...,
        help="The source file to test",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    function_name: str = Argument(..., help="The function to test"),
    output_root_dir: Optional[Path] = Option(
        None,
        help="The root directory of the output test files",
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
    ask_for_test_name: bool = Option(True, help="Allow test names to be modified"),
    ask_for_description: bool = Option(True, help="Allow adding a test description"),
    config_file: Optional[Path] = Option(
        None,
        help="The config file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=False,
    ),
):
    """
    Generate a new test.

    :param config_file:
    :param ask_for_description:
    :param ask_for_test_name:
    :param source_file:
    :param function_name:
    :param output_root_dir:
    :return:
    """
    return create_test(
        source_file=source_file,
        function_name=function_name,
        output_root_dir=output_root_dir,
        ask_for_test_name=ask_for_test_name,
        ask_for_description=ask_for_description,
        config_file_path=config_file,
    )


@app.command()
@exception_handler
def delete(
    scribe_file_path: Path = Argument(
        ...,
        help="The testscribe file to modify",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
    test_name: str = Argument("", help="The name of the test to be removed"),
):
    """
    Delete a test. This will delete the test from both the testscribe file
    and the unit test file.

    :param scribe_file_path:
    :param test_name:
    :return:
    """
    return delete_test(scribe_file_path=scribe_file_path, test_name=test_name)


@app.command()
@exception_handler
def sync(
    scribe_file_path: Path = Argument(
        ...,
        help="The testscribe file to sync",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
):
    """
    Regenerate the unit test file.

    :param scribe_file_path:
    :return:
    """
    return regenerate_tests(scribe_file_path=scribe_file_path)


@app.command()
@exception_handler
def sync_all(
    output_root_dir: Optional[Path] = Option(
        None,
        help="The root directory of the output test files",
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
):
    """
    Regenerate all unit test files under the generated test root.
    """
    return regenerate_all_tests(output_root_dir)


@app.command()
@exception_handler
def update(
    scribe_file_path: Path = Argument(
        ...,
        help="The testscribe file to update",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
    test_name: str = Argument("", help="The name of the test to update"),
) -> int:
    """
    Update the selected test.

    :param test_name:
    :param scribe_file_path:
    :return:
    """
    # The user defined setup function may call patch.
    # make sure the mode is set up before the setup function is called.
    global_var.g_test_generating_mode = True
    return update_test_cmd(scribe_file_path=scribe_file_path, test_name=test_name)


@app.command()
@exception_handler
def move(
    source_file: Path = Argument(
        ...,
        help="The source file that contains the symbol",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    class_or_function_name: str = Argument(
        ...,
        help="The name of the function or class that has moved",
    ),
    output_root_dir: Optional[Path] = Option(
        None,
        help="The root directory of the output test files",
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
):
    """
    Move tests for a function to their new files corresponding to the new module
    to which the function has moved.
    """
    return move_tests(
        source_file=source_file,
        class_or_function_name=class_or_function_name,
        output_root_path=output_root_dir,
    )


if __name__ == "__main__":  # pragma: no cover
    app()



from typing import Optional, List, Any

from testscribe.model_type import (
    ExpressionModel,
    MockModel,
    TestModel,
    PatchModel,
    ObjectModel,
    CallableModel,
)
from testscribe.namedvalues import NamedValues
from testscribe.reflection_util import get_module_str
from testscribe.util import flattern_list, BUILTIN_MODULE_NAME
from testscribe.value_input_util import import_modules_from_expression


def is_expression(v):
    return isinstance(v, ExpressionModel)


def gather_expressions_from_named_values(
    nv: Optional[NamedValues],
) -> List[ExpressionModel]:
    if nv is None:
        return []
    return [v for _, v in nv.as_list() if is_expression(v)]


def gather_expressions_from_mock_attributes(m: MockModel) -> List[ExpressionModel]:
    return [attrib for attrib in m.attributes.values() if is_expression(attrib)]


def gather_expressions_from_mock_call_return(m: MockModel) -> List[ExpressionModel]:
    return [c.return_value for c in m.calls if is_expression(c.return_value)]


def gather_expressions_from_a_mock(m: MockModel) -> List[ExpressionModel]:
    return gather_expressions_from_mock_attributes(
        m
    ) + gather_expressions_from_mock_call_return(m)


def gather_expressions_from_mocks(t: TestModel) -> List[ExpressionModel]:
    list_of_list = [gather_expressions_from_a_mock(m) for m in t.mocks]
    return flattern_list(list_of_list)


def gather_expressions_from_patches(patches: List[PatchModel]) -> List[ExpressionModel]:
    return [p.replacement for p in patches if is_expression(p.replacement)]


def gather_expressions(t: TestModel) -> List[ExpressionModel]:
    return (
        gather_expressions_from_mocks(t)
        + gather_expressions_from_named_values(t.init_parameters)
        + gather_expressions_from_named_values(t.parameters)
        + gather_expressions_from_patches(t.patches)
    )


def gather_modules_from_expressions(expressions: List[ExpressionModel]) -> List[str]:
    module_names = []
    for exp in expressions:
        _, names = import_modules_from_expression(exp.expression)
        module_names.extend(names)
    return module_names


def get_module_names_from_expressions(t: TestModel):
    """
    Can't rely on global variables to gather the module names needed to support
    input expressions
    since the global variables are only available for the current test.

    :param t:
    :return:
    """
    expressions = gather_expressions(t)
    return gather_modules_from_expressions(expressions)


def get_module_names_in_mock_specs(test: TestModel) -> List[str]:
    return [get_module_str(m.spec_str) for m in test.mocks]


def get_module_names_from_value(v: Any) -> List[str]:
    t = type(v)
    names = []
    items = []
    if t is ObjectModel:
        names.append(get_module_str(v.type))
        items = v.members.values()
    elif t is CallableModel:
        names.append(v.module)
    elif t in (list, tuple, set):
        items = v
    elif t is dict:
        items = list(v.values()) + list(v.keys())

    for i in items:
        names.extend(get_module_names_from_value(i))

    return names


def get_module_names_in_result(test: TestModel) -> List[str]:
    exception = test.exception
    if exception:
        return [get_module_str(exception.type)]
    else:
        return get_module_names_from_value(test.result)


def get_module_names_in_one_mock_call_param(params: NamedValues) -> List[str]:
    return flattern_list([get_module_names_from_value(v) for _, v in params.as_list()])


def get_module_names_in_mock_call_params(t: TestModel) -> List[str]:
    names_list = [
        get_module_names_in_one_mock_call_param(c.parameters)
        for m in t.mocks
        for c in m.calls
    ]
    names = flattern_list(names_list)
    return names


def get_referenced_modules_in_a_test(t: TestModel) -> List[str]:
    return flattern_list(
        [
            get_module_names_in_mock_specs(t),
            get_module_names_in_mock_call_params(t),
            get_module_names_in_result(t),
            get_module_names_from_expressions(t=t),
        ]
    )


def gather_import_statements_from_module_names(module_names: List[str]):
    # Conver to a set first to remove duplicates
    sorted_module_names = sorted(set(module_names))
    return [
        f"import {name}"
        for name in sorted_module_names
        if name and name != BUILTIN_MODULE_NAME
    ]


def gather_import_statements_for_referenced_modules(
    tests: List[TestModel],
) -> List[str]:
    module_names_lists = [get_referenced_modules_in_a_test(t) for t in tests]
    module_names = flattern_list(module_names_lists)
    return gather_import_statements_from_module_names(module_names)
